/*
 * Copyright 2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.jeffknecht.dynadao.api;

import java.lang.reflect.Method;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * Provides the ability to override the default behavior and map a value 
 * in the resultset to any arbitrary value.  Classes that implement
 * this interface can be set as the columnMapper value on a 
 * Column annotation.
 * 
 * In addition to implementing this interface, ColumnMapper implementations
 * must provide a default constructor and must be thread-safe.
 * 
 * @author jeff.knecht
 */
public abstract class ColumnMapper<K, V> {

	public abstract K convertToDTO(V value);
	public abstract V convertToJdbc(K value);
	
	public Class<?> getJdbcClass() {
		Method[] methods = this.getClass().getMethods();
		for (Method m : methods) {
			if (m.getName().equals("convertToJdbc") && m.getParameterTypes().length == 1) {
				return m.getReturnType();
			}
		}
		return null;
	}
	
	public Class<?> getDTOClass() {
		Method[] methods = this.getClass().getMethods();
		for (Method m : methods) {
			if (m.getName().equals("convertToDTO") && m.getParameterTypes().length == 1) {
				return m.getReturnType();
			}
		}
		return null;
	}

}
