/*
 * Copyright 2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.jeffknecht.dynadao.api;

import java.util.Collection;
import java.util.List;
import java.util.Map;


public interface DynaDao {

	/**
	 * Returns all instances of the supplied class.
	 * 
	 * @param clazz May not be null.  Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	<T> List<T> findAll(Class<T> clazz);

	/**
	 * <p>Returns all instances of the supplied class.</p>
	 * 
	 * <p>The orderBy clause should be expressed using the
	 * property names defined in the supplied clazz.  These
	 * properties should be surrounded by {curlybraces}
	 * so that they can be replaced by their proper field
	 * name.</p>
	 * 
	 * @param clazz May not be null.  Must be annotated with a Table and at least one Column definition.
	 * @param orderBy Order By clause.  May not be null.
	 * @return
	 */
	<T> List<T> findAll(Class<T> clazz, String orderBy);

	/**
	 * <p>Uses a NamedParameterJdbcTemplate to populate a list
	 * of objects of the specified type which match the supplied
	 * criteria.</p>
	 * 
	 * <p>The where clause should be expressed using the
	 * property names defined in the supplied clazz.  These
	 * properties should be surrounded by {curlybraces}
	 * so that they can be replaced by their proper field
	 * name.</p>
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @param where May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied where clause.
	 * @return
	 */
	<T> List<T> find(Class<T> clazz, String where,
			Map<String, ?> params);

	/**
	 * <p>Uses a NamedParameterJdbcTemplate to populate a list
	 * of objects of the specified type which match the supplied
	 * criteria.</p>
	 * 
	 * <p>The where and orderBy clauses should be expressed using the
	 * property names defined in the supplied clazz.  These
	 * properties should be surrounded by {curlybraces}
	 * so that they can be replaced by their proper field
	 * name.</p>
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @param where May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied where clause.
	 * @param orderBy May not be null.
	 * @return
	 */
	<T> List<T> find(Class<T> clazz, String where,
			Map<String, ?> params, String orderBy);

	/**
	 * <p>Uses a NamedParameterJdbcTemplate to locate a collection of
	 *  objects of the specified type using the supplied SQL.</p>
	 * 
	 * @param clazz May not be null. Must be annotated with at least one Column definition.
	 * @param sql May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied sql.
	 * @return A list of objects of the specified class.  If no data matches, an empty list is returned
	 */
	<T> List<T> findBySql(Class<T> clazz, final String sql, Map<String, ?> params);

	/**
	 * <p>Uses a NamedParameterJdbcTemplate to locate a single
	 * of object of the specified type which match the supplied
	 * criteria.</p>
	 * 
	 * <p>The where clause should be expressed using the
	 * property names defined in the supplied clazz.  These
	 * properties should be surrounded by [brackets]
	 * so that they can be replaced by their proper field
	 * name.</p>
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @param where May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied where clause.
	 * @return May return null if no matching records are found.
	 * @throws IncorrectResultSizeDataAccessException if more than one matching record is found
	 */
	<T> T findOne(Class<T> clazz, final String where,
			Map<String, ?> params);

	/**
	 * <p>Uses a NamedParameterJdbcTemplate to locate a single
	 * of object of the specified type using the supplied SQL.</p>
	 * 
	 * @param clazz May not be null. Must be annotated with at least one Column definition.
	 * @param sql May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied sql.
	 * @return May return null if no matching records are found.
	 * @throws IncorrectResultSizeDataAccessException if more than one matching record is found
	 */
	<T> T findOneBySql(Class<T> clazz, final String sql, Map<String, ?> params);

	/**
	 * Inserts the supplied object.  If an auto-increment column is
	 * specified in the object's class definition, it is updated with
	 * the value generated by the database.
	 * 
	 * @param object May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return The number of records updated.
	 */
	int insert(Object object);

	/**
	 * Batch inserts the supplied collection of objects.  If an auto-increment column is
	 * specified in the object's class definition, it WILL NOT be updated with
	 * the value generated by the database; this is due to the fact that the JDBC specification
	 * does not guarantee that all drivers will support this for batch operations.
	 * 
	 * @param objects May not be null. Must be annotated with a Table and at least one Column definition.  All objects in the collection must be of the same DTO class.
	 * @return The number of records updated.
	 */
	int insert(Collection<?> objects);
	
	/**
	 * <p>Updates the supplied object.  The object must have at least one column
	 * annotated as a primary key field.</p>
	 * 
	 * <p>Columns marked as optimisticLocking will be used to ensure that only
	 * records that have not been updated since the current object was retrieved
	 * can be updated.  If this test fails, then an OptimisticLockingException 
	 * is thrown</p>
	 * 
	 * @param object May not be null. Must be annotated with a Table and at least one Column definition must be marked as a primary key field.
	 * @return The number of records updated.
	 */
	int update(Object object);

	/**
	 * <p>Updates the supplied objects.  The object must have at least one column
	 * annotated as a primary key field.</p>
	 * 
	 * <p>Columns marked as optimisticLocking will be used to ensure that only
	 * records that have not been updated since the current object was retrieved
	 * can be updated.  If this test fails, then an OptimisticLockingException 
	 * is thrown</p>
	 * 
	 * @param objects May not be null. Must be annotated with a Table and at least one Column definition must be marked as a primary key field. All objects in the collection must be of the same DTO class.
	 * @return The number of records updated.
	 */
	int update(Collection<?> objects);
	
	/**
	 * <p>Deletes the supplied object.  The object must have at least one column
	 * annotated as a primary key field.</p>
	 * 
	 * <p>Columns marked as optimisticLocking will be used to ensure that only
	 * records that have not been updated since the current object was retrieved
	 * can be updated.  If this test fails, then an OptimisticLockingException 
	 * is thrown</p>
	 * 
	 * @param object May not be null. Must be annotated with a Table and at least one Column definition must be marked as a primary key field.
	 * @return The number of records deleted.
	 */
	int delete(Object object);

	/**
	 * <p>Deletes the supplied objects.  The object must have at least one column
	 * annotated as a primary key field.</p>
	 * 
	 * <p>Columns marked as optimisticLocking will be used to ensure that only
	 * records that have not been updated since the current object was retrieved
	 * can be updated.  If this test fails, then an OptimisticLockingException 
	 * is thrown</p>
	 * 
	 * @param objects May not be null. Must be annotated with a Table and at least one Column definition must be marked as a primary key field. All objects in the collection must be of the same DTO class.
	 * @return The number of records deleted.
	 */
	int delete(Collection<?> objects);

	/**
	 * <p>Uses a NamedParameterJdbcTemplate to update data using the supplied SQL.</p>
	 * 
	 * @param sql May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied sql.
	 * @return The number of records updated
	 */
	int updateBySql(final String sql, Map<String, ?> params);

	/**
	 * Counts all instances of the supplied class.
	 * 
	 * @param clazz May not be null.  Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	int countAll(Class<?> clazz);

	/**
	 * <p>Uses a count objects of the specified type which match the supplied
	 * criteria.</p>
	 * 
	 * <p>The where clause should be expressed using the
	 * property names defined in the supplied clazz.  These
	 * properties should be surrounded by {curlybraces}
	 * so that they can be replaced by their proper field
	 * name.</p>
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @param where May not be null.
	 * @param params May not be null. Must contain property values dictated by the supplied where clause.
	 * @return
	 */
	int count(Class<?> clazz, final String where,
			Map<String, ?> params);

}