/*
 * Copyright 2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.jeffknecht.dynadao.impl;

import java.lang.reflect.Field;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.jdbc.core.namedparam.AbstractSqlParameterSource;
import org.springframework.jdbc.core.namedparam.BeanPropertySqlParameterSource;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;

import com.jeffknecht.dynadao.annotation.Column;
import com.jeffknecht.dynadao.api.ColumnMapper;
import com.jeffknecht.dynadao.api.DynamicDaoDataAccessException;

/**
 * Encapsulates an object as an SqlParameterSource and extends the
 * source by allowing additional values to be set.  Note that the
 * bean values always take precedence in the case of a name
 * conflict.
 * 
 * @author jeff.knecht
 *
 */
public class AugmentedBeanParameterSource extends AbstractSqlParameterSource {
	private static final Log log = LogFactory.getLog(AugmentedBeanParameterSource.class);
	
	private Object bean;
	private Class beanClass;
	private DTOHelper dtoHelper;
	private MapSqlParameterSource internalMap;
	private BeanPropertySqlParameterSource internalBean;
	
	/**
	 * Constructor.
	 * 
	 * @param object The bean to use as the primary parameter source
	 */
	public AugmentedBeanParameterSource(Object object) {
		this.bean = object;
		this.beanClass = object.getClass();
		this.dtoHelper = DTOHelper.get(object.getClass());
		this.internalBean = new BeanPropertySqlParameterSource(object);
		this.internalMap = new MapSqlParameterSource();
		
		for(String fieldName : dtoHelper.optimisticLockingFieldNames()) {
			Object fieldValue = null;
			Field field = dtoHelper.field(fieldName);
			field.setAccessible(true);
			try {
				fieldValue = field.get(object);
			} catch (IllegalArgumentException e) {
				throw new DynamicDaoDataAccessException("Illegal argument for field " + fieldName, e);
			} catch (IllegalAccessException e) {
				throw new DynamicDaoDataAccessException("Illegal access to field " + fieldName, e);
			}
			
			int sqlType = dtoHelper.column(fieldName).type(); 
			if (sqlType != 0) {
				addValue(fieldName + "_optlock", fieldValue, sqlType);
			} else {
				addValue(fieldName + "_optlock", fieldValue);				
			}
		}
	}

	/**
	 * Indicates whether parameter name exists.
	 */
	public boolean hasValue(String paramName) {
		boolean value = internalBean.hasValue(paramName) || internalMap.hasValue(paramName);
		return value;
	}

	/**
	 * Returns the value associated with the supplied parameter name.
	 */
	public Object getValue(String paramName) {
		if (internalBean.hasValue(paramName)) {
			Field f = this.dtoHelper.field(paramName);
			Column columnAnnotation = f.getAnnotation(Column.class);
			if (columnAnnotation != null) {
				Class<? extends ColumnMapper> columnMapperClass = columnAnnotation.columnMapper();
				if (!columnMapperClass.equals(ColumnMapper.class)) {
					try {
						ColumnMapper converter = columnMapperClass.newInstance();
						f.setAccessible(true);
						return converter.convertToJdbc(f.get(this.bean));
					} catch (InstantiationException e) {
						log.error("Unable to instantiate column mapper " + columnMapperClass, e);
					} catch (IllegalAccessException e) {
						log.error("Illegal access exception attempting to instantiate column mapper " + columnMapperClass, e);
					}
				}
			}
			return internalBean.getValue(paramName);
		}
		
		if (internalMap.hasValue(paramName)) {
			return internalMap.getValue(paramName);
		}
		throw new IllegalArgumentException("No value registered for key '" + paramName + "'");
	}

	/**
	 * Adds the specified value.
	 * 
	 * @param paramName
	 * @param value
	 * @return
	 */
	public AugmentedBeanParameterSource addValue(String paramName, Object value) {
		internalMap.addValue(paramName, value);
		return this;
	}

	/**
	 * Adds the specified value.
	 * 
	 * @param paramName
	 * @param value
	 * @param sqlType
	 * @return
	 */
	public AugmentedBeanParameterSource addValue(String paramName, Object value, int sqlType) {
		internalMap.addValue(paramName, value, sqlType);
		return this;
	}

	/**
	 * Adds the specified value.
	 * 
	 * @param paramName
	 * @param value
	 * @param sqlType
	 * @param typeName
	 * @return
	 */
	public AugmentedBeanParameterSource addValue(String paramName, Object value, int sqlType, String typeName) {
		internalMap.addValue(paramName, value, sqlType, typeName);
		return this;
	}

	@Override
	public String toString() {
		if (internalBean == null && internalMap == null) {
			return super.toString();
		}
		
		StringBuilder stringBuilder = new StringBuilder("{");
		if (internalBean != null) {
			String[] beanParameterNames = internalBean.getReadablePropertyNames();
			for(String paramName : beanParameterNames) {
				if (stringBuilder.length() > 0) {
					stringBuilder.append(" ");
				}
				stringBuilder.append("[" + paramName + "] = [" + internalBean.getValue(paramName) + "]");
			}
		}
		if (internalMap != null) {
			Map<String, Object> values = internalMap.getValues();
			for (Map.Entry<String, Object> entry : values.entrySet()) {
				if (stringBuilder.length() > 0) {
					stringBuilder.append(" ");
				}
				stringBuilder.append("[" + entry.getKey() + "] = [" + entry.getValue() + "]");
			}
		}
		stringBuilder.append("}");
		return stringBuilder.toString();
	}
}
