/*
 * Copyright 2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.jeffknecht.dynadao.impl;

import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.dao.EmptyResultDataAccessException;
import org.springframework.dao.OptimisticLockingFailureException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.BeanPropertySqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.jdbc.support.GeneratedKeyHolder;

import com.jeffknecht.dynadao.annotation.Column;
import com.jeffknecht.dynadao.annotation.Select;
import com.jeffknecht.dynadao.api.DynaDao;
import com.jeffknecht.dynadao.api.DynamicDaoDataAccessException;
import com.jeffknecht.dynadao.api.Generator;
import com.jeffknecht.dynadao.api.Generator.Operation;


/**
 * Dynamic data access object.  Provides the ability to 
 * query and updated database tables based on annotations in the
 * DTO class definition.
 * 
 * Note that the intention here is not to replace JPA or Hibernate.
 * These frameworks are very good at what they do, but introduce some
 * interesting complexities with respect to session management.  Rather,
 * the intention of this library is to simplify the 80% usage scenario
 * of simple CRUD operations against a single table.
 * 
 * @author jeff.knecht
 *
 */
public class DynaDaoImpl implements DynaDao {
	
	private static final Log log = LogFactory.getLog(DynaDaoImpl.class);
	
	private static final int MAX_CACHE_SIZE = 256;
	
	private static final HashMap<Class<? extends Generator>, Generator> GENERATORS = new LinkedHashMap<Class<? extends Generator>, Generator>(MAX_CACHE_SIZE, 0.75f) {
		@Override
		protected boolean removeEldestEntry(
				java.util.Map.Entry<Class<? extends Generator>, Generator> eldest) {
			return size() > MAX_CACHE_SIZE;
		}
	};
	
	private final NamedParameterJdbcTemplate namedParameterjdbcTemplate;
	private final JdbcTemplate jdbcTemplate;
	private final DataSource dataSource;
	
	/**
	 * Constructor.
	 *  
	 * @param ds The DataSource to use for executing data access calls
	 */
	public DynaDaoImpl(DataSource ds) {
		this.dataSource = ds;
		this.jdbcTemplate = new JdbcTemplate(ds);
		this.namedParameterjdbcTemplate = new NamedParameterJdbcTemplate(ds);
	}
	
	/**
	 * Constructor for unit and performance tests that want to simulate
	 * or interrogate calls to the jdbcTemplate classes.
	 * 
	 * @param namedParameterJdbcTemplate
	 * @param jdbcTemplate
	 */
	protected DynaDaoImpl(NamedParameterJdbcTemplate namedParameterJdbcTemplate, JdbcTemplate jdbcTemplate) {
		this.dataSource = null;
		this.namedParameterjdbcTemplate = namedParameterJdbcTemplate;
		this.jdbcTemplate = jdbcTemplate;
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#findAll(java.lang.Class)
	 */
	public <T> List<T> findAll(Class<T> clazz) {
		if (log.isDebugEnabled()) {
			log.debug("findAll called with class: " + clazz);
		}
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		if (selectAnnotation != null) {
			if (selectAnnotation.where().length() > 0) {
				sql.append(" ");
				sql.append(whereClauseSQL(clazz, selectAnnotation.where()));				
			}
			if (selectAnnotation.groupBy().length() > 0) {
				sql.append(" ");
				sql.append(groupByClauseSQL(clazz, selectAnnotation.groupBy()));				
			}
			if (selectAnnotation.having().length() > 0) {
				sql.append(" ");
				sql.append(havingClauseSQL(clazz, selectAnnotation.having()));				
			}
			if (selectAnnotation.orderBy().length() > 0) {
				sql.append(" ");
				sql.append(orderByClauseSQL(clazz, selectAnnotation.orderBy()));				
			}
		}
		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sql.toString());
		}
		return jdbcTemplate.query(sql.toString(), new DynaRowMapper<T>(clazz), (Object[]) null);
	}

	public <T> List<T> findAll(Class<T> clazz, String orderBy) {
		if (log.isDebugEnabled()) {
			log.debug("findAll called with class: " + clazz + ", orderBy: " + orderBy);
		}
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		StringBuilder orderByBuilder = new StringBuilder(orderBy.trim());
		if (selectAnnotation != null) {
			if (selectAnnotation.where().length() > 0) {
				sql.append(" ");
				sql.append(whereClauseSQL(clazz, selectAnnotation.where()));				
			}
			if (selectAnnotation.groupBy().length() > 0) {
				sql.append(" ");
				sql.append(groupByClauseSQL(clazz, selectAnnotation.groupBy()));				
			}
			if (selectAnnotation.having().length() > 0) {
				sql.append(" ");
				sql.append(havingClauseSQL(clazz, selectAnnotation.having()));				
			}
			if (selectAnnotation.orderBy().length() > 0) {
				if (orderByBuilder.length() > 0) {
					orderByBuilder.append(", ");
				}
				orderByBuilder.append(selectAnnotation.orderBy());
			}
		}
		
		sql.append(" ");
		sql.append(orderByClauseSQL(clazz, orderByBuilder.toString()));
		
		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sql.toString());
		}
		return jdbcTemplate.query(sql.toString(), new DynaRowMapper<T>(clazz), (Object[]) null);
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#find(java.lang.Class, java.lang.String, java.util.Map)
	 */
	public <T> List<T> find(Class<T> clazz, String where, Map<String, ?> params) {
		if (log.isDebugEnabled()) {
			log.debug("find called with class: " + clazz + ", where: " + where + ", params: " + buildStringRepresentationOfParameters(params));
		}
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		StringBuilder whereBuilder = new StringBuilder(where.trim());
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		if (selectAnnotation != null) {
			if (selectAnnotation.where().length() > 0) {
				if (whereBuilder.length() > 0) {
					whereBuilder.append(" and ");
				}
				whereBuilder.append(selectAnnotation.where());				
			}
		}
		
		sql.append(" ");
		sql.append(whereClauseSQL(clazz, whereBuilder.toString()));
		
		if (selectAnnotation != null) {
			if (selectAnnotation.groupBy().length() > 0) {
				sql.append(" ");
				sql.append(groupByClauseSQL(clazz, selectAnnotation.groupBy()));				
			}
			if (selectAnnotation.having().length() > 0) {
				sql.append(" ");
				sql.append(havingClauseSQL(clazz, selectAnnotation.having()));				
			}
			if (selectAnnotation.orderBy().length() > 0) {
				sql.append(" ");
				sql.append(orderByClauseSQL(clazz, selectAnnotation.orderBy()));				
			}
		}

		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sql.toString());
			log.debug("Parameters: " + buildStringRepresentationOfParameters(params));
		}
		return namedParameterjdbcTemplate.query(sql.toString(), params, new DynaRowMapper<T>(clazz));
	}

	public <T> List<T> find(Class<T> clazz, String where,
			Map<String, ?> params, String orderBy) {
		if (log.isDebugEnabled()) {
			log.debug("find called with class: " + clazz + ", where: " + where +", params: " + buildStringRepresentationOfParameters(params) + ", orderBy: " + orderBy);
		}
		
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		
		StringBuilder whereBuilder = new StringBuilder(where.trim());
		StringBuilder orderByBuilder = new StringBuilder(orderBy.trim());
		
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		if (selectAnnotation != null) {
			if (selectAnnotation.where().length() > 0) {
				if (whereBuilder.length() > 0) {
					whereBuilder.append(" and ");
				}
				whereBuilder.append(selectAnnotation.where());				
			}
		}
		
		sql.append(" ");
		sql.append(whereClauseSQL(clazz, whereBuilder.toString()));
		
		if (selectAnnotation != null) {
			if (selectAnnotation.groupBy().length() > 0) {
				sql.append(" ");
				sql.append(groupByClauseSQL(clazz, selectAnnotation.groupBy()));				
			}
			if (selectAnnotation.having().length() > 0) {
				sql.append(" ");
				sql.append(havingClauseSQL(clazz, selectAnnotation.having()));				
			}
			if (selectAnnotation.orderBy().length() > 0) {
				if (orderByBuilder.length() > 0) {
					orderByBuilder.append(", ");
				}
				orderByBuilder.append(selectAnnotation.orderBy());
			}
		}

		sql.append(" ");
		sql.append(orderByClauseSQL(clazz, orderByBuilder.toString()));
				
		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sql.toString());
			log.debug("Parameters: " + buildStringRepresentationOfParameters(params));
		}
		return namedParameterjdbcTemplate.query(sql.toString(), params, new DynaRowMapper<T>(clazz));
	}

	public <T> List<T> findBySql(Class<T> clazz, String sql,
			Map<String, ?> params) {
		if (log.isDebugEnabled()) {
			log.debug("findBySql called with class: " + clazz);
			log.debug("SQL Statement: " + sql.toString());
			log.debug("Parameters: " + buildStringRepresentationOfParameters(params));
		}
		return namedParameterjdbcTemplate.query(sql, params, new DynaRowMapper<T>(clazz));
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#find(java.lang.Class, java.lang.String, java.util.HashMap)
	 */
	public <T> T findOne(Class<T> clazz, String where, Map<String, ?> params) {
		if (log.isDebugEnabled()) {
			log.debug("findOne called with class: " + clazz + ", where: " + where + ", params: " + buildStringRepresentationOfParameters(params));
		}
		StringBuilder sql = new StringBuilder(baseSelectSQL(clazz));
		
		StringBuilder whereBuilder = new StringBuilder(where.trim());
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		if (selectAnnotation != null) {
			if (selectAnnotation.where().length() > 0) {
				if (whereBuilder.length() > 0) {
					whereBuilder.append(" and ");
				}
				whereBuilder.append(selectAnnotation.where());				
			}
		}
		
		sql.append(" ");
		sql.append(whereClauseSQL(clazz, whereBuilder.toString()));
		
		if (selectAnnotation != null) {
			if (selectAnnotation.groupBy().length() > 0) {
				sql.append(" ");
				sql.append(groupByClauseSQL(clazz, selectAnnotation.groupBy()));				
			}
			if (selectAnnotation.having().length() > 0) {
				sql.append(" ");
				sql.append(havingClauseSQL(clazz, selectAnnotation.having()));				
			}
			if (selectAnnotation.orderBy().length() > 0) {
				sql.append(" ");
				sql.append(orderByClauseSQL(clazz, selectAnnotation.orderBy()));				
			}
		}

		try {
			if (log.isDebugEnabled()) {
				log.debug("Generated SQL Statement: " + sql.toString());
				log.debug("Parameters: " + buildStringRepresentationOfParameters(params));
			}
			return namedParameterjdbcTemplate.queryForObject(sql.toString(), params, new DynaRowMapper<T>(clazz));
		} catch (EmptyResultDataAccessException ignored) {
			return null;
		}
	}

	public <T> T findOneBySql(Class<T> clazz, String sql, Map<String, ?> params) {
		try {
			if (log.isDebugEnabled()) {
				log.debug("findOneBySql called with class: " + clazz);
				log.debug("Supplied SQL Statement: " + sql);
				log.debug("Parameters: " + buildStringRepresentationOfParameters(params));
			}
			return namedParameterjdbcTemplate.queryForObject(sql.toString(), params, new DynaRowMapper<T>(clazz));
		} catch (EmptyResultDataAccessException ignored) {
			return null;
		}
	}
	
	private String buildStringRepresentationOfParameters(
			Map<String, ?> params) {
		StringBuilder paramStringBuilder = new StringBuilder();
		for (Map.Entry<String, ?> entry : params.entrySet()) {
			if (paramStringBuilder.length() > 0) {
				paramStringBuilder.append(" ");
			}
			paramStringBuilder.append("{[" + entry.getKey() + "] = [" + entry.getValue() + "]}");
		}
		return paramStringBuilder.toString();
	}

	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#insert(java.lang.Object)
	 */
	public int insert(Object object) {
		if (log.isDebugEnabled()) {
			log.debug("insert called with object: " + object);
		}
		Select selectAnnotation = object.getClass().getAnnotation(Select.class);
		if (selectAnnotation != null) {
			throw new UnsupportedOperationException("insert is not supported for objects defined with the @Select annotation");
		}
		
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		
		String sql = insertSQL(object.getClass());

		//generate the column values
		for (String fieldName : dtoHelper.fieldNames()) {
			Field field = dtoHelper.field(fieldName);
			Column column = field.getAnnotation(Column.class);
			Class<? extends Generator> generatorClass = column.generator();
			if (column.generateOnInsert() && !(generatorClass == null) && !(generatorClass.equals(Generator.class))) {
				executeGenerator(object, field, column, generatorClass, Operation.INSERT);
			}
		}
		
		if (dtoHelper.autoIncrementFieldName() == null) {
			SqlParameterSource params = new AugmentedBeanParameterSource(object);
			if (log.isDebugEnabled()) {
				log.debug("Generated SQL Statement: " + sql);
				log.debug("Parameters: " + params);
			}
			return namedParameterjdbcTemplate.update(sql, params);
		} else {
			SqlParameterSource params = new AugmentedBeanParameterSource(object);
			String fieldName = dtoHelper.autoIncrementFieldName();
			GeneratedKeyHolder keyHolder = new GeneratedKeyHolder();

			if (log.isDebugEnabled()) {
				log.debug("Generated SQL Statement: " + sql);
				log.debug("Parameters: " + params);
			}
			int returnValue = namedParameterjdbcTemplate.update(sql, params, keyHolder);
			
			Number key = keyHolder.getKey();
			Field field = dtoHelper.field(fieldName);
			field.setAccessible(true);
			try {
				if (log.isDebugEnabled()) {
					log.debug("generated key value " + key.longValue() + " is being stored in field " + field.getName());
				}
				Class<?> fieldType = field.getType();
				if (fieldType.equals(int.class)) {
					field.setInt(object, key.intValue());
				} else if (fieldType.equals(Integer.class)) {
					field.set(object, Integer.valueOf(key.intValue()));
				} else if (fieldType.equals(long.class)) {
					field.setLong(object, key.longValue());
				} else if (fieldType.equals(Long.class)) {
					field.set(object, Long.valueOf(key.longValue()));
				} else if (fieldType.equals(short.class)) {
					field.setShort(object, key.shortValue());
				} else if (fieldType.equals(Short.class)) {
					field.set(object, Short.valueOf(key.shortValue()));
				} else if (fieldType.equals(byte.class)) {
					field.setByte(object, key.byteValue());
				} else if (fieldType.equals(Byte.class)) {
					field.set(object, Byte.valueOf(key.byteValue()));
				} else if (fieldType.equals(double.class)) {
					field.setDouble(object, key.doubleValue());
				} else if (fieldType.equals(Double.class)) {
					field.set(object, Double.valueOf(key.doubleValue()));
				} else if (fieldType.equals(float.class)) {
					field.setFloat(object, key.floatValue());
				} else if (fieldType.equals(Float.class)) {
					field.set(object, Float.valueOf(key.floatValue()));
				} else if (fieldType.equals(BigInteger.class)) {
					field.set(object, BigInteger.valueOf(key.longValue()));
				} else if (fieldType.equals(BigDecimal.class)) {
					field.set(object, BigDecimal.valueOf(key.doubleValue()));
				} else {
					throw new UnsupportedOperationException("Data type " + fieldType + " is not supported for autoincrement fields.");
				}
			} catch (IllegalAccessException e) {
				throw new DynamicDaoDataAccessException("Illegal access on field " + field, e);
			}
			return returnValue;
		}
	}
	
	public int insert(Collection<?> objects) {
		if (log.isDebugEnabled()) {
			log.debug("insert called with object: " + objects);
		}
		if (objects.size() == 0) {
			return 0;
		}
		
		String sql = null;
		DTOHelper dtoHelper = null;
		SqlParameterSource[] paramArray = new SqlParameterSource[objects.size()];
		int i = 0;
		for (Object object : objects) {
			if (sql == null) {
				Select selectAnnotation = object.getClass().getAnnotation(Select.class);
				if (selectAnnotation != null) {
					throw new UnsupportedOperationException("insert is not supported for objects defined with the @Select annotation");
				}
				
				dtoHelper = DTOHelper.get(object.getClass());
				
				sql = insertSQL(object.getClass());
			}
			
			//generate the column values
			for (String fieldName : dtoHelper.fieldNames()) {
				Field field = dtoHelper.field(fieldName);
				Column column = field.getAnnotation(Column.class);
				Class<? extends Generator> generatorClass = column.generator();
				if (column.generateOnInsert() && !(generatorClass == null) && !(generatorClass.equals(Generator.class))) {
					executeGenerator(object, field, column, generatorClass, Operation.INSERT);
				}
			}
			
			SqlParameterSource params = new AugmentedBeanParameterSource(object);
			if (log.isDebugEnabled()) {
				log.debug("Generated SQL Statement: " + sql);
				log.debug("Parameters: " + params);
			}
			paramArray[i] = params;
			i++;
		}
		
		int[] inserts = namedParameterjdbcTemplate.batchUpdate(sql, paramArray);
		int insertCount = 0;
		for (int insert : inserts) {
			insertCount += insert;
		}
		return insertCount;
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#update(java.lang.Object)
	 */
	public int update(Object object) {
		if (log.isDebugEnabled()) {
			log.debug("update called with object: " + object);
		}
		Select selectAnnotation = object.getClass().getAnnotation(Select.class);
		if (selectAnnotation != null) {
			throw new UnsupportedOperationException("update is not supported for objects defined with the @Select annotation");
		}
		
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		Set<String> primaryKeyFieldNames = dtoHelper.primaryKeyFieldNames();
		if (primaryKeyFieldNames.size() == 0) {
			throw new DynamicDaoDataAccessException("Unable to uniquely identify object " + object + ", because it has no primary key fields");
		}

		
		AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);
		StringBuilder sqlBuilder = new StringBuilder(baseUpdateSQL(object.getClass()));
		sqlBuilder.append(" where ");
		sqlBuilder.append(computeWhereClause(object, dtoHelper, paramSource,
				primaryKeyFieldNames));
		
		//execute generators after the optimistic locking fields have been identified
		for (String fieldName : dtoHelper.fieldNames()) {
			Field field = dtoHelper.field(fieldName);
			Column column = field.getAnnotation(Column.class);
			Class<? extends Generator> generatorClass = column.generator();
			if (column.generateOnUpdate() && !(generatorClass == null) && !(generatorClass.equals(Generator.class))) {
				executeGenerator(object, field, column, generatorClass, Operation.UPDATE);
			}
		}
		
		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sqlBuilder.toString());
			log.debug("Parameters: " + paramSource);
		}
		int updated = namedParameterjdbcTemplate.update(sqlBuilder.toString(), paramSource);
		if (updated == 0 && dtoHelper.optimisticLockingFieldNames().size() > 0) {
			throw new OptimisticLockingFailureException("Update failed.  No record exists with either the supplied primary key or optimistic locking column values.");
		}
		return updated;
	}

	public int update(Collection<?> objects) {
		if (log.isDebugEnabled()) {
			log.debug("update called with objects: " + objects);
		}
		if (objects.size() == 0) {
			return 0;
		}
		
		String sql = null;
		DTOHelper dtoHelper = null;
		SqlParameterSource[] paramArray = new SqlParameterSource[objects.size()];
		int i = 0;
		for (Object object : objects) {
			if (sql == null) {
				Select selectAnnotation = object.getClass().getAnnotation(Select.class);
				if (selectAnnotation != null) {
					throw new UnsupportedOperationException("update is not supported for objects defined with the @Select annotation");
				}
				
				dtoHelper = DTOHelper.get(object.getClass());
				Set<String> primaryKeyFieldNames = dtoHelper.primaryKeyFieldNames();
				if (primaryKeyFieldNames.size() == 0) {
					throw new DynamicDaoDataAccessException("Unable to uniquely identify object of type " + object.getClass() + ", because it has no primary key fields");
				}
		
				
				AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);
				StringBuilder sqlBuilder = new StringBuilder(baseUpdateSQL(object.getClass()));
				sqlBuilder.append(" where ");
				sqlBuilder.append(computeWhereClause(object, dtoHelper, paramSource,
						primaryKeyFieldNames));
				sql = sqlBuilder.toString();
				
				if (log.isDebugEnabled()) {
					log.debug("Generated SQL Statement: " + sql);
				}
			}

			//do this before running the generator to ensure the optimistic locking parameters are saved before being modified
			AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);

			//execute generators after the optimistic locking fields have been identified
			for (String fieldName : dtoHelper.fieldNames()) {
				Field field = dtoHelper.field(fieldName);
				Column column = field.getAnnotation(Column.class);
				Class<? extends Generator> generatorClass = column.generator();
				if (column.generateOnUpdate() && !(generatorClass == null) && !(generatorClass.equals(Generator.class))) {
					executeGenerator(object, field, column, generatorClass, Operation.UPDATE);
				}
			}

			if (log.isDebugEnabled()) {
				log.debug("Parameters: " + paramSource);
			}
			paramArray[i] = paramSource;
			i++;
		}
		int[] updated = namedParameterjdbcTemplate.batchUpdate(sql, paramArray);
		int updatedCount = 0;
		for (int update : updated) {
			updatedCount += update;
		}
		if (updatedCount < objects.size() && dtoHelper.optimisticLockingFieldNames().size() > 0) {
			throw new OptimisticLockingFailureException("Update failed.  No record exists with either the supplied primary key or optimistic locking column values.");
		}
		return updatedCount;
	}
	
	/* (non-Javadoc)
	 * @see hyatt.eas.dao.dynadao.DynaDao#delete(java.lang.Object)
	 */
	public int delete(Object object) {
		if (log.isDebugEnabled()) {
			log.debug("delete called with object: " + object);
		}
		Select selectAnnotation = object.getClass().getAnnotation(Select.class);
		if (selectAnnotation != null) {
			throw new UnsupportedOperationException("delete is not supported for objects defined with the @Select annotation");
		}
		
		DTOHelper dtoHelper = DTOHelper.get(object.getClass());
		AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);
		Set<String> primaryKeyFieldNames = dtoHelper.primaryKeyFieldNames();
		if (primaryKeyFieldNames.size() == 0) {
			throw new DynamicDaoDataAccessException("Unable to uniquely identify object " + object + ", because it has no primary key fields");
		}
		StringBuilder sqlBuilder = new StringBuilder(baseDeleteSQL(object.getClass()));
		sqlBuilder.append(" where ");
		sqlBuilder.append(computeWhereClause(object, dtoHelper, paramSource,
				primaryKeyFieldNames));
		
		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sqlBuilder.toString());
			log.debug("Parameters: " + paramSource);
		}
		int updated = namedParameterjdbcTemplate.update(sqlBuilder.toString(), paramSource);
		if (updated == 0 && dtoHelper.optimisticLockingFieldNames().size() > 0) {
			throw new OptimisticLockingFailureException("Delete failed.  No record exists with either the supplied primary key or optimistic locking column values.");
		}
		return updated;
	}

	public int delete(Collection<?> objects) {
		if (log.isDebugEnabled()) {
			log.debug("delete called with objects: " + objects);
		}
		if (objects.size() == 0) {
			return 0;
		}
		
		String sql = null;
		DTOHelper dtoHelper = null;
		SqlParameterSource[] paramArray = new SqlParameterSource[objects.size()];
		int i = 0;
		for (Object object : objects) {
			if (sql == null) {
				Select selectAnnotation = object.getClass().getAnnotation(Select.class);
				if (selectAnnotation != null) {
					throw new UnsupportedOperationException("delete is not supported for objects defined with the @Select annotation");
				}
				
				dtoHelper = DTOHelper.get(object.getClass());
				Set<String> primaryKeyFieldNames = dtoHelper.primaryKeyFieldNames();
				if (primaryKeyFieldNames.size() == 0) {
					throw new DynamicDaoDataAccessException("Unable to uniquely identify object of type " + object.getClass() + ", because it has no primary key fields");
				}
		
				
				AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);
				StringBuilder sqlBuilder = new StringBuilder(baseDeleteSQL(object.getClass()));
				sqlBuilder.append(" where ");
				sqlBuilder.append(computeWhereClause(object, dtoHelper, paramSource,
						primaryKeyFieldNames));
				sql = sqlBuilder.toString();
				
				if (log.isDebugEnabled()) {
					log.debug("Generated SQL Statement: " + sql);
				}
			}

			//do this before running the generator to ensure the optimistic locking parameters are saved before being modified
			AugmentedBeanParameterSource paramSource = new AugmentedBeanParameterSource(object);

			//execute generators after the optimistic locking fields have been identified
			for (String fieldName : dtoHelper.fieldNames()) {
				Field field = dtoHelper.field(fieldName);
				Column column = field.getAnnotation(Column.class);
				Class<? extends Generator> generatorClass = column.generator();
				if (column.generateOnUpdate() && !(generatorClass == null) && !(generatorClass.equals(Generator.class))) {
					executeGenerator(object, field, column, generatorClass, Operation.UPDATE);
				}
			}

			if (log.isDebugEnabled()) {
				log.debug("Parameters: " + paramSource);
			}
			paramArray[i] = paramSource;
			i++;
		}
		int[] deleted = namedParameterjdbcTemplate.batchUpdate(sql, paramArray);
		int deletedCount = 0;
		for (int delete : deleted) {
			deletedCount += delete;
		}
		if (deletedCount < objects.size() && dtoHelper.optimisticLockingFieldNames().size() > 0) {
			throw new OptimisticLockingFailureException("Delete failed.  No record exists with either the supplied primary key or optimistic locking column values.");
		}
		return deletedCount;
	}
	
	public int updateBySql(String sql, Map<String, ?> params) {
		if (log.isDebugEnabled()) {
			log.debug("updateBySql called.");
			log.debug("Supplied SQL Statement: " + sql);
			log.debug("Parameters: " + buildStringRepresentationOfParameters(params));
		}
		return namedParameterjdbcTemplate.update(sql, params);
	}
	
	public int countAll(Class<?> clazz) {
		if (log.isDebugEnabled()) {
			log.debug("countAll called with class: " + clazz);
		}
		
		StringBuilder sql = new StringBuilder();
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		if (selectAnnotation != null) {
			sql.append("select count(*) from (");
			sql.append(baseSelectSQL(clazz));
			if (selectAnnotation.where().length() > 0) {
				sql.append(" ");
				sql.append(whereClauseSQL(clazz, selectAnnotation.where()));				
			}
			if (selectAnnotation.groupBy().length() > 0) {
				sql.append(" ");
				sql.append(groupByClauseSQL(clazz, selectAnnotation.groupBy()));				
			}
			if (selectAnnotation.having().length() > 0) {
				sql.append(" ");
				sql.append(havingClauseSQL(clazz, selectAnnotation.having()));				
			}
			if (selectAnnotation.orderBy().length() > 0) {
				sql.append(" ");
				sql.append(orderByClauseSQL(clazz, selectAnnotation.orderBy()));				
			}
			sql.append(")");
		} else {
			sql.append(baseCountSQL(clazz));
		}
		
		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sql.toString());
		}
		return jdbcTemplate.queryForInt(sql.toString(), (Object[]) null);
	}

	public int count(Class<?> clazz, String where, Map<String, ?> params) {
		if (log.isDebugEnabled()) {
			log.debug("count called with class: " + clazz + ", where: " + where + ", params: " + buildStringRepresentationOfParameters(params));
		}
		
		StringBuilder sql = new StringBuilder();
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		StringBuilder whereBuilder = new StringBuilder(where.trim());
		if (selectAnnotation != null) {
			sql.append("select count(*) from (");
			sql.append(baseSelectSQL(clazz));
			if (selectAnnotation.where().length() > 0) {
				if (whereBuilder.length() > 0) {
					whereBuilder.append(" and ");
				}
				whereBuilder.append(selectAnnotation.where());				
			}
		} else {
			sql.append(baseCountSQL(clazz));
		}
		
		sql.append(" ");
		sql.append(whereClauseSQL(clazz, whereBuilder.toString()));
		
		if (selectAnnotation != null) {
			if (selectAnnotation.groupBy().length() > 0) {
				sql.append(" ");
				sql.append(groupByClauseSQL(clazz, selectAnnotation.groupBy()));				
			}
			if (selectAnnotation.having().length() > 0) {
				sql.append(" ");
				sql.append(havingClauseSQL(clazz, selectAnnotation.having()));				
			}
			if (selectAnnotation.orderBy().length() > 0) {
				sql.append(" ");
				sql.append(orderByClauseSQL(clazz, selectAnnotation.orderBy()));				
			}
			sql.append(")");
		}
		
		if (log.isDebugEnabled()) {
			log.debug("Generated SQL Statement: " + sql);
			log.debug("Parameters: " + buildStringRepresentationOfParameters(params));
		}
		return namedParameterjdbcTemplate.queryForInt(sql.toString(), params);
	}

	private String whereClauseSQL(Class<?> clazz, String where) {
		StringBuilder whereClauseBuilder = new StringBuilder();
		if (where.trim().length() > 0 && !where.trim().toLowerCase().startsWith("where ")) {
			whereClauseBuilder.append("where ");
		}
		whereClauseBuilder.append(clauseSQL(clazz, where));
		return whereClauseBuilder.toString();
	}

	private String groupByClauseSQL(Class<?> clazz, String clause) {
		StringBuilder clauseBuilder = new StringBuilder();
		if (clause.trim().length() > 0 && !clause.trim().toLowerCase().startsWith("group by ")) {
			clauseBuilder.append("group by ");
		}
		clauseBuilder.append(clauseSQL(clazz, clause));
		return clauseBuilder.toString();
	}

	private String havingClauseSQL(Class<?> clazz, String clause) {
		StringBuilder clauseBuilder = new StringBuilder();
		if (clause.trim().length() > 0 && !clause.trim().toLowerCase().startsWith("having ")) {
			clauseBuilder.append("having ");
		}
		clauseBuilder.append(clauseSQL(clazz, clause));
		return clauseBuilder.toString();
	}

	private String orderByClauseSQL(Class<?> clazz, String orderBy) {
		StringBuilder orderByBuilder = new StringBuilder();
		if (orderBy.trim().length() > 0 && !orderBy.trim().toLowerCase().startsWith("order by ")) {
			orderByBuilder.append("order by ");
		}
		orderByBuilder.append(clauseSQL(clazz, orderBy));
		return orderByBuilder.toString();
	}
	
	private String clauseSQL(Class<?> clazz, String clause) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		Pattern p = Pattern.compile("\\[[^\\[\\]]*\\]");	// match any text surrounded by {curly braces}
															// except ignore curly braces inside of curly braces
		
		String targetClause = clause.trim();
		Matcher matcher = p.matcher(targetClause);
		while (matcher.find()) {
			String toReplace = matcher.group();
			String fieldName = toReplace.substring(1, toReplace.length() - 1);
			String columnName = resolveColumnName(dtoHelper, fieldName);
			targetClause = targetClause.replace(toReplace, columnName);
			matcher = p.matcher(targetClause);
		}

		return targetClause;
	}
	

	private void executeGenerator(Object object, Field field, Column column,
			Class<? extends Generator> generatorClass, Operation operation) {

		field.setAccessible(true);
		try {
			Generator generator = GENERATORS.get(generatorClass);
			if (generator == null) {
				generator = generatorClass.newInstance();
				GENERATORS.put(generatorClass, generator);
			}
			generator.generateValue(field, object, operation);
		} catch (InstantiationException e) {
			throw new DynamicDaoDataAccessException("Unable to instantiate generator " + column.generator(), e);
		} catch (IllegalAccessException e) {
			throw new DynamicDaoDataAccessException("Unable to instantiate generator " + column.generator(), e);
		}
	}

	private String computeWhereClause(Object object, DTOHelper dtoHelper,
			AugmentedBeanParameterSource paramSource,
			Set<String> primaryKeyFieldNames) {
		
		StringBuilder sqlBuilder = new StringBuilder();
		int whereCounter = 0;
		for(String fieldName : primaryKeyFieldNames) {
			if (whereCounter > 0) {
				sqlBuilder.append(" and ");
			}
			sqlBuilder.append(dtoHelper.columnName(fieldName)).append(" = :").append(fieldName);
			whereCounter++;
		}
		
		for(String fieldName : dtoHelper.optimisticLockingFieldNames()) {
			if (whereCounter > 0) {
				sqlBuilder.append(" and ");
			}
			sqlBuilder.append(dtoHelper.columnName(fieldName)).append(" = :").append(fieldName + "_optlock");
			
			whereCounter++;
		}
		
		return sqlBuilder.toString();
	}

	/**
	 * Generates a count statement for the supplied class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table.
	 * @return
	 */
	protected String baseCountSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		return "select count(*) from " + dtoHelper.table();
	}

	/**
	 * Generates a select statement for the supplied class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String baseSelectSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		Select selectAnnotation = clazz.getAnnotation(Select.class);
		if (selectAnnotation == null) {
			return "select " + selectColumnList(dtoHelper) + " from " + dtoHelper.table();
		} else {
			return "select " + selectColumnList(dtoHelper) + " from " + selectAnnotation.from();
		}
	}

	/**
	 * Generates a sql insert statement for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String insertSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		
		StringBuilder sb = new StringBuilder("insert into ");
		sb.append(dtoHelper.table());
		sb.append(" (");
		int columnCount = 0;
		
		StringBuilder fieldBuilder = new StringBuilder();
		for (String fieldName : dtoHelper.fieldNames()) {
			Column column = dtoHelper.column(fieldName);
			if (!column.autoIncrement()) {
				if (columnCount > 0) {
					sb.append(", ");
					fieldBuilder.append(", ");
				}
				sb.append(column.value());
				fieldBuilder.append(":").append(fieldName);
				columnCount++;
			}
		}
		sb.append(") values (").append(fieldBuilder.toString()).append(")");
		return sb.toString();
	}

	/**
	 * Generates an update statement without the WHERE clause
	 * for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String baseUpdateSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sb = new StringBuilder("update ");
		sb.append(dtoHelper.table());
		sb.append(" set ");
		int columnCount = 0;
		
		for (String fieldName : dtoHelper.fieldNames()) {
			Column column = dtoHelper.column(fieldName);
			if (!column.autoIncrement()) {
				if (columnCount > 0) {
					sb.append(", ");
				}
				sb.append(column.value()).append(" = :").append(fieldName);
				columnCount++;
			}
		}
		return sb.toString();
	}

	/**
	 * Generates a delete statement without the WHERE clause
	 * for the supplied DTO class.
	 * 
	 * @param clazz May not be null. Must be annotated with a Table and at least one Column definition.
	 * @return
	 */
	protected String baseDeleteSQL(Class<?> clazz) {
		DTOHelper dtoHelper = DTOHelper.get(clazz);
		StringBuilder sb = new StringBuilder("delete from ");
		sb.append(dtoHelper.table());
		return sb.toString();
	}

	/**
	 * Generates a comma-delimited list of column names based
	 * on the columns provided by the dtoHelper.
	 * 
	 * @param dtoHelper
	 * @return
	 */
	protected String selectColumnList(DTOHelper dtoHelper) {
		StringBuilder sb = new StringBuilder();
		for (Column column : dtoHelper.columns()) {
			if (sb.length() > 0) {
				sb.append(", ");
			}
			if (column.derivedFrom().length() > 0) {
				sb.append(column.derivedFrom()).append(" as ");
			}
			sb.append(column.value());			
		}
		return sb.toString();
	}

	/**
	 * Translates a field name into a database column name.
	 * 
	 * @param dtoHelper
	 * @param fieldName
	 * @return
	 */
	protected String resolveColumnName(DTOHelper dtoHelper, String fieldName) {
		return dtoHelper.columnName(fieldName);
	}

	/**
	 * Exposes the DataSource used by this object.
	 * 
	 * @return
	 */
	protected DataSource getDataSource() {
		return dataSource;
	}

}
