/*
 * Copyright 2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.jeffknecht.dynadao.impl;


import java.lang.reflect.Field;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Date;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.HashMap;
import java.util.LinkedHashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.support.lob.DefaultLobHandler;
import org.springframework.jdbc.support.lob.LobHandler;

import com.jeffknecht.dynadao.annotation.Column;
import com.jeffknecht.dynadao.api.ColumnMapper;
import com.jeffknecht.dynadao.api.DynamicDaoDataAccessException;

/**
 * Dynamically maps records in a resultset to instances of a class
 * that is annotated with a Table and at least one Column annotation.
 * The class definition must have a default constructor.
 * 
 * @author jeff.knecht
 *
 * @param <T>
 */
public class DynaRowMapper<T> implements RowMapper<T> {
	private static final Log log = LogFactory.getLog(DynaRowMapper.class);
	
	private static final int MAX_CACHE_SIZE = 256;

	private static HashMap<Class<? extends ColumnMapper>, ColumnMapper> columnMappers = new LinkedHashMap<Class<? extends ColumnMapper>, ColumnMapper>(MAX_CACHE_SIZE, 0.75f) {
		@Override
		protected boolean removeEldestEntry(
				java.util.Map.Entry<Class<? extends ColumnMapper>, ColumnMapper> eldest) {
			return size() > MAX_CACHE_SIZE;
		}
	};

	private Class<? extends T> clazz;
	private LobHandler lobHandler = new DefaultLobHandler();
	private DTOHelper dtoHelper;
	
	/**
	 * Instantiates a DynaRowMapper.  The supplied class must
	 * be the same as the type defined for this 
	 * @param clazz
	 */
	public DynaRowMapper(Class<? extends T> clazz) {
		this.clazz = clazz;
		dtoHelper = DTOHelper.get(clazz);
		if (log.isDebugEnabled()) {
			log.debug("Instantiated DynaRowMapper for class " + clazz);
		}
	}
	
	/**
	 * Maps the current row in the resultset to an instance of the
	 * underlying class definition.
	 */
	public T mapRow(ResultSet rs, int arg1) throws SQLException {
		try {
			if (log.isDebugEnabled()) {
				log.debug("Mapping row for instance of class " + clazz);
			}
			T t = (T) clazz.newInstance();
			//find the column definitions and get the value out of the resultset
			for (String fieldName : dtoHelper.fieldNames()) {
				
				Column column = dtoHelper.column(fieldName);
				String columnName = column.value();
				int columnType = column.type();
				boolean trim = column.trim();
				Field f = dtoHelper.field(fieldName);
				Class<?> fieldType = f.getType();
				f.setAccessible(true);
				try {
					Class<? extends ColumnMapper> columnMapperClass = column.columnMapper();
					ColumnMapper customMapper = null;
					if (columnMapperClass != null && !(columnMapperClass.equals(ColumnMapper.class))) {
						customMapper = columnMappers.get(columnMapperClass);
						if (customMapper == null) {
							customMapper = column.columnMapper().newInstance();
							columnMappers.put(columnMapperClass, customMapper);
						}
						fieldType = customMapper.getJdbcClass();
					}
					
					if (fieldType.equals(String.class)) {
						if (columnType == Types.CLOB) {
							if (customMapper != null) {
								f.set(t, customMapper.convertToDTO(lobHandler.getClobAsString(rs, columnName)));
							} else if (trim) {
								f.set(t, trim(lobHandler.getClobAsString(rs, columnName)));
							} else {
								f.set(t, lobHandler.getClobAsString(rs, columnName));
							}
						} else {
							if (customMapper != null) {
								f.set(t, customMapper.convertToDTO(rs.getString(columnName)));
							}else if (trim) {
								f.set(t, trim(rs.getString(columnName)));
							} else {
								f.set(t, rs.getString(columnName));
							}
						}
					} else if (fieldType.equals(int.class)) {
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(rs.getInt(columnName)));							
						} else {
							f.set(t, rs.getInt(columnName));
						}
					} else if (fieldType.equals(Integer.class)) {
						Integer value = rs.getInt(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(long.class)) {
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(rs.getLong(columnName)));							
						} else {
							f.set(t, rs.getLong(columnName));
						}
					} else if (fieldType.equals(Long.class)) {
						Long value = rs.getLong(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(short.class)) {
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(rs.getShort(columnName)));							
						} else {
							f.set(t, rs.getShort(columnName));
						}
					} else if (fieldType.equals(Short.class)) {
						Short value = rs.getShort(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(double.class)) {
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(rs.getDouble(columnName)));							
						} else {
							f.set(t, rs.getDouble(columnName));
						}
					} else if (fieldType.equals(Double.class)) {
						Double value = rs.getDouble(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(float.class)) {
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(rs.getFloat(columnName)));							
						} else {
							f.set(t, rs.getFloat(columnName));
						}
					} else if (fieldType.equals(Float.class)) {
						Float value = rs.getFloat(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(boolean.class)) {
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(rs.getBoolean(columnName)));							
						} else {
							f.set(t, rs.getBoolean(columnName));
						}
					} else if (fieldType.equals(Boolean.class)) {
						Boolean value = rs.getBoolean(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(char.class)) {
						String val = rs.getString(columnName);
						char c = val == null || val.length() == 0 ? 0 : val.charAt(0);
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(c));							
						} else {
							f.set(t, c);
						}
					} else if (fieldType.equals(Character.class)) {
						String val = rs.getString(columnName);
						Character c = val == null || val.length() == 0 ? null : val.charAt(0);
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(c));							
						} else {
							f.set(t, c);
						}
					} else if (fieldType.equals(java.util.Date.class)) {
						Timestamp value = rs.getTimestamp(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(java.sql.Date.class)) {
						Date value = rs.getDate(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(Time.class)) {
						Time value = rs.getTime(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(Timestamp.class)) {
						Timestamp value = rs.getTimestamp(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(byte[].class)) {
						byte[] value = lobHandler.getBlobAsBytes(rs, columnName);
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(BigDecimal.class)) {
						BigDecimal value = rs.getBigDecimal(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else if (fieldType.equals(BigInteger.class)) {
						Long longValue = rs.getLong(columnName);
						BigInteger value = null;
						if (!rs.wasNull()) {
							value = new BigInteger(longValue.toString());
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					} else {
						Object value = rs.getObject(columnName);
						if (rs.wasNull()) {
							value = null;
						}
						if (customMapper != null) {
							f.set(t, customMapper.convertToDTO(value));
						} else {
							f.set(t, value);
						}
					}
				} catch (IllegalAccessException e) {
					throw new DynamicDaoDataAccessException("Illegal access on field " + f, e);
				}
			}
			if (log.isDebugEnabled()) {
				log.debug("Row mapped to object: " + t);
			}
			return t;
		} catch (InstantiationException e) {
			throw new DynamicDaoDataAccessException("Unable to instantiate class " + clazz, e);
		} catch (IllegalAccessException e) {
			throw new DynamicDaoDataAccessException("Illegal access on class " + clazz, e);
		}
	}

	private String trim(String string) {
		if (string == null) {
			return null;
		} else {
			return string.trim();
		}
	}
}
